/*
 * Copyright (c) 2081, Sonos, Inc.
 *
 * SPDX-License-Identifier:     GPL-2.0
 *
 * Driver for A113 Timer peripheral
 */

#include <linux/clk.h>
#include <linux/device.h>
#include <linux/io.h>
#include <linux/module.h>
#include <linux/of_platform.h>
#include <linux/amlogic/iomap.h>
#include <linux/proc_fs.h>

#include "lla.h"
#include "blackbox.h"

#define DRV_NAME	"hrtimer_a113"

#define HRTIMER_A113_CONFIGURED_RATE	1000000
#define HRTIMER_A113_REG_BASE_ADDR		0xff800000
#define HRTIMER_A113_REG_AO_TIMERE		0x54

#define HRTIMER_A113_READ(reg)	aml_read_aobus(reg);
#define HRTIMER_A113_WRITE(reg, val)	aml_write_aobus(reg, val);
#define HRTIMER_A113_UPDATE_BITS(reg, mask, val)	aml_aobus_update_bits(reg, mask, val);

struct hrtimer_a113_priv {
	struct lla_hw_clk		lla_clk_src;
	unsigned int			rate;
	struct device			*dev;
};

static uint32_t hrtimer_a113_lla_clock_read(void *context)
{
	(void)context;
	return HRTIMER_A113_READ(HRTIMER_A113_REG_AO_TIMERE);
}

static struct lla_hw_clk_ops lla_clk_ops = {
	.read = hrtimer_a113_lla_clock_read,
};

static int hrtimer_a113_probe(struct platform_device *pdev)
{
	struct hrtimer_a113_priv *priv;
	priv = devm_kzalloc(&pdev->dev, sizeof(struct hrtimer_a113_priv), GFP_KERNEL);
	if (!priv) {
		bb_log_dev(&pdev->dev, BB_MOD_LLA, BB_LVL_ERR, "Memory allocation failed!");
		return -ENOMEM;
	}
	platform_set_drvdata(pdev, priv);
	priv->dev = &(pdev->dev);

	HRTIMER_A113_WRITE(HRTIMER_A113_REG_AO_TIMERE, 0xdeadbeef);

	if (lla_init_clk(&(priv->lla_clk_src), DRV_NAME, (HRTIMER_A113_REG_BASE_ADDR + HRTIMER_A113_REG_AO_TIMERE),
		HRTIMER_A113_CONFIGURED_RATE, 1, &lla_clk_ops, priv)) {
		bb_log_dev(&pdev->dev, BB_MOD_LLA, BB_LVL_ERR, "Could not set up LLA clock.");
		return -EFAULT;
	}

	if (lla_register_clk(&(priv->lla_clk_src))) {
		bb_log_dev(&pdev->dev, BB_MOD_LLA, BB_LVL_ERR, "Could not register LLA clock.");
		return -ENODEV;
	}

	bb_log_dev(&pdev->dev, BB_MOD_LLA, BB_LVL_INFO, "Registered hrtimer for the LLA. Rate: %u Hz", (HRTIMER_A113_CONFIGURED_RATE));

	return 0;
}

static int hrtimer_a113_remove(struct platform_device *pdev)
{
	(void)pdev;

	return 0;
}

static const struct of_device_id hrtimer_a113_ids[] = {
	{ .compatible = "sonos,hrtimer-a113", },
	{ }
};

static struct platform_driver hrtimer_a113_driver = {
	.probe = hrtimer_a113_probe,
	.remove = hrtimer_a113_remove,
	.driver = {
		.name = DRV_NAME,
		.owner = THIS_MODULE,
		.of_match_table = hrtimer_a113_ids,
	},
};

int hrtimer_a113_init(void)
{
	return platform_driver_register(&hrtimer_a113_driver);
}

void hrtimer_a113_exit(void)
{
	platform_driver_unregister(&hrtimer_a113_driver);
}
