/*
 * Copyright (c) 2018, Sonos, Inc.
 *
 * SPDX-License-Identifier:     GPL-2.0
 *
 * Driver for ALSA Shim layer.
 */
#include <linux/time.h>
#include <linux/proc_fs.h>
#include <linux/platform_device.h>
#include <linux/fs.h>
#include <linux/of.h>
#include <sound/core.h>
#include <sound/minors.h>
#include <sound/pcm.h>
#include <sound/pcm_params.h>
#include <sound/info.h>
#include <linux/soundcard.h>
#include <sound/initval.h>
#include <asm/barrier.h>

#include "lla.h"
#include "blackbox.h"
#include "alsa_shim.h"

#define DRV_NAME "alsa_lla"
#define COMPLETION_HIST_SIZE 10

#define TX_RUNNING_FLAG_CLEARED	0x1
#define RX_RUNNING_FLAG_CLEARED	0x2

struct alsa_lla {
	struct device *dev;
	struct snd_pcm *tx_pcm;
	struct snd_pcm *rx_pcm;
	struct lla_dev tx_dev;
	struct lla_dev rx_dev;
	struct lla_shared *tx_shared;
	struct lla_shared *rx_shared;
	uint32_t tx_mask;
	uint32_t rx_mask;
	u64 last_tx_frame;
	u64 last_rx_frame;
	u32 tx_completion_hist[COMPLETION_HIST_SIZE];
	u32 tx_completion_max;
	u32 rx_completion_hist[COMPLETION_HIST_SIZE];
	u32 rx_completion_max;
	wait_queue_head_t tx_poll_event;
	wait_queue_head_t rx_poll_event;
	int running_flag_cleared_on_suspend;
};

struct alsa_lla *global_data;

#define HW_PARAM(v) [SNDRV_PCM_HW_PARAM_##v] = #v
static const char * const snd_pcm_hw_param_names[] = {
	HW_PARAM(ACCESS),
	HW_PARAM(FORMAT),
	HW_PARAM(SUBFORMAT),
	HW_PARAM(SAMPLE_BITS),
	HW_PARAM(FRAME_BITS),
	HW_PARAM(CHANNELS),
	HW_PARAM(RATE),
	HW_PARAM(PERIOD_TIME),
	HW_PARAM(PERIOD_SIZE),
	HW_PARAM(PERIOD_BYTES),
	HW_PARAM(PERIODS),
	HW_PARAM(BUFFER_TIME),
	HW_PARAM(BUFFER_SIZE),
	HW_PARAM(BUFFER_BYTES),
	HW_PARAM(TICK_TIME),
};

int hacky_alsa_pcm_id_parser(char *id)
{
	int ret = -1;

	if (!strncmp(id, "TDM-A-0-dac", 11) || !strncmp(id, "TDM-B-0-dac", 11) || !strncmp(id, "TDM-C-0-dac", 11)) {
		ret = 0;
	} else if (strncmp(id, "PDM-1-mic", 9) == 0) {
		ret = 1;
	} else if (strncmp(id, "TDM-B-5-hdmi_rx", 15) == 0) {
		ret = 5;
	}

	return ret;
}


static int snd_interval_refine_set(struct snd_interval *i, unsigned int val)
{
	struct snd_interval t;

	t.empty = 0;
	t.min = t.max = val;
	t.openmin = t.openmax = 0;
	t.integer = 1;

	return snd_interval_refine(i, &t);
}


static int _snd_pcm_hw_param_set(struct snd_pcm_hw_params *params,
				 snd_pcm_hw_param_t var, unsigned int val,
				 int dir)
{
	int changed;

	if (hw_is_mask(var)) {
		struct snd_mask *m = hw_param_mask(params, var);
		if (val == 0 && dir < 0) {
			changed = -EINVAL;
			snd_mask_none(m);
		} else {
			if (dir > 0) {
				val++;
			} else if (dir < 0) {
				val--;
			}
			changed = snd_mask_refine_set(hw_param_mask(params, var), val);
		}
	} else if (hw_is_interval(var)) {
		struct snd_interval *i = hw_param_interval(params, var);
		if (val == 0 && dir < 0) {
			changed = -EINVAL;
			snd_interval_none(i);
		} else if (dir == 0) {
			changed = snd_interval_refine_set(i, val);
		} else {
			struct snd_interval t;
			t.openmin = 1;
			t.openmax = 1;
			t.empty = 0;
			t.integer = 0;
			if (dir < 0) {
				t.min = val - 1;
				t.max = val;
			} else {
				t.min = val;
				t.max = val + 1;
			}
			changed = snd_interval_refine(i, &t);
		}
	} else {
		return -EINVAL;
	}

	if (changed) {
		params->cmask |= 1 << var;
		params->rmask |= 1 << var;
	}

	return changed;
}

#if 0
void dump_hw_params(struct snd_pcm_hw_params *params)
{
	static int j = 0;
	unsigned int k;
	struct snd_interval *i = NULL;
	struct snd_mask *m = NULL;

	printk("HW PARAMS DUMP: %d\n", j++);
	for (k = SNDRV_PCM_HW_PARAM_FIRST_MASK; k <= SNDRV_PCM_HW_PARAM_LAST_MASK; k++)
	{
		m = hw_param_mask(params, k);
		if (snd_mask_empty(m)) {
			continue;
		}
		if (!(params->rmask & (1 << k))) {
			continue;
		}
		printk("M %s = ", snd_pcm_hw_param_names[k]);
		printk("%04x%04x%04x%04x \n", m->bits[3], m->bits[2], m->bits[1], m->bits[0]);
	}
	for (k = SNDRV_PCM_HW_PARAM_FIRST_INTERVAL; k <= SNDRV_PCM_HW_PARAM_LAST_INTERVAL; k++) {
		i = hw_param_interval(params, k);
		if (snd_interval_empty(i)) {
			continue;
		}
		if (!(params->rmask & (1 << k))) {
			continue;
		}
		printk("I %s = ", snd_pcm_hw_param_names[k]);
		if (i->empty) {
			printk("empty\n");
		} else {
			printk("%c%u %u%c\n", i->openmin ? '(' : '[', i->min, i->max, i->openmax ? ')' : ']');
		}
	}
}
#endif


static int alsa_lla_tx_commit(unsigned long idx, void *context)
{
	struct alsa_lla *alsa_data = context;
	if (idx >= alsa_data->tx_shared->buffer_cnt) {
		bb_log_dev(alsa_data->dev, BB_MOD_LLA, BB_LVL_ERR, "Buffer index %lu out of bounds.", idx);
		return -EFAULT;
	}
	return 0;
}

static unsigned int alsa_lla_tx_poll(struct file *filp, struct poll_table_struct *table, void *context)
{
	struct alsa_lla *alsa_data = context;
	unsigned int mask = 0;

	poll_wait(filp, &(alsa_data->tx_poll_event), table);

	if (alsa_data->tx_shared->poll_num < (alsa_data->tx_shared->buffers_complete + alsa_data->tx_shared->tx_latency)) {
		mask |= POLLOUT | POLLWRNORM;
	}

	return mask;
}

void alsa_lla_tx_complete(struct snd_pcm_substream *substream)
{
	struct alsa_lla *alsa_data = global_data;
	struct snd_pcm_runtime *runtime = substream->runtime;
	u32 completion_time = lla_clock_read();
	u64 now_frames = 0;
	u64 num_completed = 0;
	now_frames = runtime->status->hw_ptr;
	if (now_frames < alsa_data->last_tx_frame) {
		bb_log_dev(alsa_data->dev, BB_MOD_LLA, BB_LVL_ERR, "TX frame counter went backwards! now %llu last %llu", now_frames, alsa_data->last_tx_frame);
	}
	num_completed = (now_frames - alsa_data->last_tx_frame) / alsa_data->tx_shared->frames_per_buffer;
	if (num_completed < COMPLETION_HIST_SIZE) {
		alsa_data->tx_completion_hist[num_completed]++;
	} else {
		if (num_completed > alsa_data->tx_completion_max) {
			alsa_data->tx_completion_max = num_completed;
		}
		alsa_data->tx_completion_hist[COMPLETION_HIST_SIZE - 1]++;
	}
	runtime->control->appl_ptr = alsa_data->last_tx_frame;
	alsa_data->last_tx_frame += num_completed * alsa_data->tx_shared->frames_per_buffer;
	if (num_completed) {
		alsa_data->tx_shared->complete_time = completion_time;
		dmb(sy);
		alsa_data->tx_shared->buffers_complete += num_completed;
		dmb(sy);
		alsa_data->tx_shared->complete_time2 = completion_time;
	}
	wake_up_interruptible(&(alsa_data->tx_poll_event));
}

int alsa_lla_tx_open(struct file *filp, void *context)
{
	struct alsa_lla *alsa_data = context;
	struct snd_pcm_substream *substream = alsa_data->tx_pcm->streams[0].substream;
	struct snd_pcm_runtime *runtime;
	struct snd_pcm_hw_params params;
	struct snd_pcm_sw_params sw = {
		.tstamp_mode = SNDRV_PCM_TSTAMP_NONE,
		.period_step = 1,
		.avail_min = 128,
		.start_threshold = 0,
		.stop_threshold = 4096,
		.silence_threshold = 4096,
		.silence_size = 128,
	};
	struct lla_ptr lla_buf;
	int ret;

	bb_log_dev(alsa_data->dev, BB_MOD_LLA, BB_LVL_DEBUG, "tx open");
	ret = snd_pcm_open_substream(alsa_data->tx_pcm, SNDRV_PCM_STREAM_PLAYBACK, filp, &substream);
	if (ret) {
		bb_log_dev(alsa_data->dev, BB_MOD_LLA, BB_LVL_ERR, "snd_pcm_open_substream returned %d", ret);
		return ret;
	}
	_snd_pcm_hw_params_any(&params);
	_snd_pcm_hw_param_set(&params, SNDRV_PCM_HW_PARAM_ACCESS, SNDRV_PCM_ACCESS_MMAP_INTERLEAVED, 0);
	_snd_pcm_hw_param_set(&params, SNDRV_PCM_HW_PARAM_PERIOD_SIZE, 128, 0);
	_snd_pcm_hw_param_set(&params, SNDRV_PCM_HW_PARAM_PERIODS, 32, 0);
	ret = snd_pcm_kernel_ioctl(substream, SNDRV_PCM_IOCTL_HW_PARAMS, &params);
	if (ret) {
		bb_log_dev(alsa_data->dev, BB_MOD_LLA, BB_LVL_ERR, "pcm_ioctl_hw_params returned %d", ret);
		return ret;
	}

	bb_log_dev(alsa_data->dev, BB_MOD_LLA, BB_LVL_DEBUG, "ACCESS %x", params_access(&params));
	bb_log_dev(alsa_data->dev, BB_MOD_LLA, BB_LVL_DEBUG, "FORMAT %x", params_format(&params));
	bb_log_dev(alsa_data->dev, BB_MOD_LLA, BB_LVL_DEBUG, "SUBFORMAT %x", params_subformat(&params));
	bb_log_dev(alsa_data->dev, BB_MOD_LLA, BB_LVL_DEBUG, "CHANNELS %d", params_channels(&params));
	bb_log_dev(alsa_data->dev, BB_MOD_LLA, BB_LVL_DEBUG, "RATE %d", params_rate(&params));
	bb_log_dev(alsa_data->dev, BB_MOD_LLA, BB_LVL_DEBUG, "PERIODS %d", params_periods(&params));
	bb_log_dev(alsa_data->dev, BB_MOD_LLA, BB_LVL_DEBUG, "PERIOD_SIZE %d", params_period_size(&params));
	bb_log_dev(alsa_data->dev, BB_MOD_LLA, BB_LVL_DEBUG, "BUFFER_SIZE %d", params_buffer_size(&params));
	bb_log_dev(alsa_data->dev, BB_MOD_LLA, BB_LVL_DEBUG, "BUFFER_BYTES %d", params_buffer_bytes(&params));

	ret = snd_pcm_kernel_ioctl(substream, SNDRV_PCM_IOCTL_SW_PARAMS, &sw);
	if (ret) {
		bb_log_dev(alsa_data->dev, BB_MOD_LLA, BB_LVL_ERR, "pcm_ioctl_sw_params returned %d", ret);
		return ret;
	}
	runtime = substream->runtime;
	lla_buf = (struct lla_ptr){
		.k = runtime->dma_buffer_p->area,
		.p = runtime->dma_buffer_p->addr
	};
	memset(lla_buf.k, 0x00, runtime->dma_bytes);
	bb_log_dev(alsa_data->dev, BB_MOD_LLA, BB_LVL_DEBUG, "alloc tx buf %p/%lx clen %u cpf %u fpb %lu bufs %lu runtime buf %lu per %lu dma %lu",
				lla_buf.k, lla_buf.p,
				runtime->sample_bits/8,
				runtime->channels,
				runtime->period_size,
				runtime->buffer_size / runtime->period_size,
				runtime->buffer_size,
				runtime->period_size,
				runtime->dma_bytes);
	alsa_data->tx_shared = lla_allocate_shared_buffers(&(alsa_data->tx_dev), runtime->sample_bits/8, runtime->channels, runtime->period_size, runtime->buffer_size / runtime->period_size, 1, &lla_buf);
	lla_update_sample_rate(&(alsa_data->tx_dev), runtime->rate);
	alsa_data->tx_shared->tx_latency = 10;
	alsa_data->tx_shared->channel_mask = alsa_data->tx_mask;
	alsa_data->last_tx_frame = 0;
	alsa_data->tx_shared->flags |= LLA_SHARED_FLAG_TX | LLA_SHARED_FLAG_RUNNING;

	runtime->stop_threshold = runtime->boundary;

	ret = snd_pcm_kernel_ioctl(substream, SNDRV_PCM_IOCTL_PREPARE, NULL);
	if (ret) {
		bb_log_dev(alsa_data->dev, BB_MOD_LLA, BB_LVL_ERR, "pcm_ioctl_prepare returned %d", ret);
	}
	ret = snd_pcm_kernel_ioctl(substream, SNDRV_PCM_IOCTL_START, NULL);
	if (ret) {
		bb_log_dev(alsa_data->dev, BB_MOD_LLA, BB_LVL_ERR, "pcm_ioctl_start returned %d", ret);
	}

	runtime->ext_completion_cb = alsa_lla_tx_complete;

	return 0;
}

int alsa_lla_tx_release(struct file *filp, void *context)
{
	struct alsa_lla *alsa_data = context;
	struct snd_pcm_substream *substream = alsa_data->tx_pcm->streams[0].substream;
	substream->runtime->ext_completion_cb = NULL;
	alsa_data->tx_shared->buffers.k = NULL;
	lla_free_shared_buffers(&(alsa_data->tx_dev));
	alsa_data->tx_shared = NULL;
	snd_pcm_release_substream(substream);
	return 0;
}

static struct lla_ops alsa_lla_tx_ops = {
	.commit		= alsa_lla_tx_commit,
	.poll		= alsa_lla_tx_poll,
	.attach		= alsa_lla_tx_open,
	.release	= alsa_lla_tx_release,
};

void alsa_lla_rx_complete(struct snd_pcm_substream *substream)
{
	struct alsa_lla *alsa_data = global_data;
	struct snd_pcm_runtime *runtime = substream->runtime;
	u32 completion_time = lla_clock_read();
	u64 now_frames = 0;
	u64 num_completed = 0;
	now_frames = runtime->status->hw_ptr;
	if (now_frames < alsa_data->last_rx_frame) {
		bb_log_dev(alsa_data->dev, BB_MOD_LLA, BB_LVL_ERR, "RX frame counter went backwards! now %llu last %llu", now_frames, alsa_data->last_rx_frame);
	}
	num_completed = (now_frames - alsa_data->last_rx_frame) / alsa_data->rx_shared->frames_per_buffer;
	if (num_completed < COMPLETION_HIST_SIZE) {
		alsa_data->rx_completion_hist[num_completed]++;
	} else {
		if (num_completed > alsa_data->rx_completion_max) {
			alsa_data->rx_completion_max = num_completed;
		}
		alsa_data->rx_completion_hist[COMPLETION_HIST_SIZE - 1]++;
	}
	runtime->control->appl_ptr = alsa_data->last_rx_frame;
	alsa_data->last_rx_frame += num_completed * alsa_data->rx_shared->frames_per_buffer;
	if (num_completed) {
		alsa_data->rx_shared->complete_time = completion_time;
		dmb(sy);
		alsa_data->rx_shared->buffers_complete += num_completed;
		dmb(sy);
		alsa_data->rx_shared->complete_time2 = completion_time;
	}
	wake_up_interruptible(&(alsa_data->rx_poll_event));
}

unsigned int alsa_lla_rx_poll(struct file *filp, struct poll_table_struct *table, void *context)
{
	struct alsa_lla *alsa_data = context;
	u32 mask = 0;

	poll_wait(filp, &(alsa_data->rx_poll_event), table);

	if ((alsa_data->rx_shared->poll_num < alsa_data->rx_shared->buffers_complete) && (alsa_data->rx_shared->flags & LLA_SHARED_FLAG_RUNNING)) {
		mask |= POLLIN | POLLRDNORM;
	}

	return mask;
}

int alsa_lla_rx_open(struct file *filp, void *context)
{
	struct alsa_lla *alsa_data = context;
	struct snd_pcm_substream *substream = alsa_data->rx_pcm->streams[1].substream;
	struct snd_pcm_runtime *runtime;
	struct snd_pcm_hw_params params;
	struct snd_pcm_sw_params sw = {
		.tstamp_mode = SNDRV_PCM_TSTAMP_NONE,
		.period_step = 1,
		.avail_min = 1024,
		.start_threshold = 0,
		.stop_threshold = 4096,
		.silence_threshold = 0,
		.silence_size = 0,
	};
	struct lla_ptr lla_buf;
	int ret;

	bb_log_dev(alsa_data->dev, BB_MOD_LLA, BB_LVL_DEBUG, "rx open");
	ret = snd_pcm_open_substream(alsa_data->rx_pcm, SNDRV_PCM_STREAM_CAPTURE, filp, &substream);
	if (ret) {
		bb_log_dev(alsa_data->dev, BB_MOD_LLA, BB_LVL_ERR, "snd_pcm_open_substream returned %d", ret);
		return ret;
	}
	_snd_pcm_hw_params_any(&params);
	_snd_pcm_hw_param_set(&params, SNDRV_PCM_HW_PARAM_ACCESS, SNDRV_PCM_ACCESS_MMAP_INTERLEAVED, 0);
	_snd_pcm_hw_param_set(&params, SNDRV_PCM_HW_PARAM_PERIOD_SIZE, 128, 0);
	_snd_pcm_hw_param_set(&params, SNDRV_PCM_HW_PARAM_PERIODS, 32, 0);

	ret = snd_pcm_kernel_ioctl(substream, SNDRV_PCM_IOCTL_HW_PARAMS, &params);
	if (ret) {
		bb_log_dev(alsa_data->dev, BB_MOD_LLA, BB_LVL_ERR, "pcm_ioctl_hw_params returned %d", ret);
		return ret;
	}

	bb_log_dev(alsa_data->dev, BB_MOD_LLA, BB_LVL_DEBUG, "ACCESS %x", params_access(&params));
	bb_log_dev(alsa_data->dev, BB_MOD_LLA, BB_LVL_DEBUG, "FORMAT %x", params_format(&params));
	bb_log_dev(alsa_data->dev, BB_MOD_LLA, BB_LVL_DEBUG, "SUBFORMAT %x", params_subformat(&params));
	bb_log_dev(alsa_data->dev, BB_MOD_LLA, BB_LVL_DEBUG, "CHANNELS %d", params_channels(&params));
	bb_log_dev(alsa_data->dev, BB_MOD_LLA, BB_LVL_DEBUG, "RATE %d", params_rate(&params));
	bb_log_dev(alsa_data->dev, BB_MOD_LLA, BB_LVL_DEBUG, "PERIODS %d", params_periods(&params));
	bb_log_dev(alsa_data->dev, BB_MOD_LLA, BB_LVL_DEBUG, "PERIOD_SIZE %d", params_period_size(&params));
	bb_log_dev(alsa_data->dev, BB_MOD_LLA, BB_LVL_DEBUG, "BUFFER_SIZE %d", params_buffer_size(&params));
	bb_log_dev(alsa_data->dev, BB_MOD_LLA, BB_LVL_DEBUG, "BUFFER_BYTES %d", params_buffer_bytes(&params));

	ret = snd_pcm_kernel_ioctl(substream, SNDRV_PCM_IOCTL_SW_PARAMS, &sw);
	if (ret) {
		bb_log_dev(alsa_data->dev, BB_MOD_LLA, BB_LVL_ERR, "pcm_ioctl_sw_params returned %d", ret);
		return ret;
	}
	runtime = substream->runtime;
	lla_buf = (struct lla_ptr){
		.k = runtime->dma_buffer_p->area,
		.p = runtime->dma_buffer_p->addr
	};
	bb_log_dev(alsa_data->dev, BB_MOD_LLA, BB_LVL_DEBUG, "alloc rx buf %p/%lx clen %u cpf %u fpb %lu bufs %lu runtime buf %lu per %lu dma %lu",
				lla_buf.k, lla_buf.p,
				runtime->sample_bits/8,
				runtime->channels,
				runtime->period_size,
				runtime->buffer_size / runtime->period_size,
				runtime->buffer_size,
				runtime->period_size,
				runtime->dma_bytes);
	alsa_data->rx_shared = lla_allocate_shared_buffers(&(alsa_data->rx_dev), runtime->sample_bits/8, runtime->channels, runtime->period_size, runtime->buffer_size / runtime->period_size, 1, &lla_buf);
	lla_update_sample_rate(&(alsa_data->rx_dev), runtime->rate);
	alsa_data->rx_shared->channel_mask = alsa_data->rx_mask;
	alsa_data->last_rx_frame = 0;

	runtime->stop_threshold = runtime->boundary;

	ret = snd_pcm_kernel_ioctl(substream, SNDRV_PCM_IOCTL_PREPARE, NULL);
	if (ret) {
		bb_log_dev(alsa_data->dev, BB_MOD_LLA, BB_LVL_ERR, "pcm_ioctl_prepare returned %d", ret);
	}
	ret = snd_pcm_kernel_ioctl(substream, SNDRV_PCM_IOCTL_START, NULL);
	if (ret) {
		bb_log_dev(alsa_data->dev, BB_MOD_LLA, BB_LVL_ERR, "pcm_ioctl_start returned %d", ret);
	}

	runtime->ext_completion_cb = alsa_lla_rx_complete;
	alsa_data->rx_shared->flags |= LLA_SHARED_FLAG_RX | LLA_SHARED_FLAG_RUNNING;

	return 0;
}

int alsa_lla_rx_release(struct file *filp, void *context)
{
	struct alsa_lla *alsa_data = context;
	struct snd_pcm_substream *substream = alsa_data->rx_pcm->streams[1].substream;
	substream->runtime->ext_completion_cb = NULL;
	alsa_data->rx_shared->buffers.k = NULL;
	lla_free_shared_buffers(&(alsa_data->rx_dev));
	alsa_data->rx_shared = NULL;
	snd_pcm_release_substream(substream);
	return 0;
}

static struct lla_ops alsa_lla_rx_ops = {
	.poll		= alsa_lla_rx_poll,
	.attach		= alsa_lla_rx_open,
	.release	= alsa_lla_rx_release,
};

static int alsa_lla_register(struct snd_pcm *pcm)
{
	int i;
	if (!pcm) {
		bb_log(BB_MOD_LLA, BB_LVL_ERR, "No PCM struct!");
		return -1;
	}
	for (i = 0; i < 2; i++) {
		if (pcm->streams[i].substream) {
			switch (hacky_alsa_pcm_id_parser(pcm->id)) {
				case 0:
					global_data->tx_pcm = pcm;
					lla_register_dev(&(global_data->tx_dev));
					bb_log(BB_MOD_LLA, BB_LVL_INFO, "ALSA device '%s' Registered.", pcm->id);
					break;
				case 1:
					global_data->rx_pcm = pcm;
					lla_register_dev(&(global_data->rx_dev));
					bb_log(BB_MOD_LLA, BB_LVL_INFO, "ALSA device '%s' Registered.", pcm->id);
					break;
				case 2:
					break;
				case 3:
					break;
				case 5:
					break;
				default:
					bb_log(BB_MOD_LLA, BB_LVL_INFO, "Ignoring unknown device '%s'", pcm->id);
					break;
			}
		}
	}
	return 0;
}

static int alsa_lla_unregister(struct snd_pcm *pcm)
{
	int i;
	if (!pcm) {
		bb_log(BB_MOD_LLA, BB_LVL_ERR, "No PCM struct!");
		return -1;
	}
	for (i = 0; i < 2; i++) {
		if (pcm->streams[i].substream) {
			switch (hacky_alsa_pcm_id_parser(pcm->id)) {
				case 0:
					global_data->tx_pcm = NULL;
					lla_unregister_dev(&(global_data->tx_dev));
					bb_log(BB_MOD_LLA, BB_LVL_INFO, "ALSA device '%s' Unregistered.", pcm->id);
					break;
				case 1:
					global_data->rx_pcm = NULL;
					lla_unregister_dev(&(global_data->rx_dev));
					bb_log(BB_MOD_LLA, BB_LVL_INFO, "ALSA device '%s' Unregistered.", pcm->id);
					break;
				case 2:
					break;
				case 3:
					break;
				case 5:
					break;
				default:
					bb_log(BB_MOD_LLA, BB_LVL_INFO, "Ignoring unknown device '%s'", pcm->id);
					break;
			}
		}
	}
	return 0;
}

static int alsa_lla_disconnect(struct snd_pcm *pcm)
{
	return 0;
}

static struct snd_pcm_notify alsa_lla_notify =
{
	.n_register	= alsa_lla_register,
	.n_disconnect	= alsa_lla_disconnect,
	.n_unregister	= alsa_lla_unregister,
};


#ifdef SNOOP_RX
static int alsa_procfs_snoop_show(struct seq_file *m, void *v)
{
	struct alsa_lla *alsa_data = m->private;
	seq_write(m, alsa_data->rx_shared->buffers.k, alsa_data->rx_shared->buffers_total);
	return 0;
}

static int alsa_procfs_snoop_open(struct inode *inode, struct file *file)
{
	return single_open(file, alsa_procfs_snoop_show, PDE_DATA(inode));
}

static const struct file_operations procfs_snoop_alsa_ops = {
	.owner = THIS_MODULE,
	.open = alsa_procfs_snoop_open,
	.read = seq_read,
	.llseek = seq_lseek,
	.release = single_release,
};
#endif

static int alsa_procfs_show(struct seq_file *m, void *v)
{
	struct alsa_lla *alsa_data = m->private;
	struct snd_pcm_runtime *tx_runtime = alsa_data->tx_pcm->streams[0].substream->runtime;
	struct snd_pcm_runtime *rx_runtime = alsa_data->rx_pcm->streams[1].substream->runtime;
	int idx;
	if (tx_runtime) {
		seq_printf(m, "Current TX DMA location in frames: %08lu\n", tx_runtime->status->hw_ptr);
	}
	seq_printf(m, "Last completed TX frame: %08llu\n", alsa_data->last_tx_frame);
	seq_printf(m, "TX Buffer completion histogram:\n");
	for (idx = 0; idx < COMPLETION_HIST_SIZE; idx++) {
		if (idx == COMPLETION_HIST_SIZE - 1) {
			seq_printf(m, "%u+: %u\n", idx, alsa_data->tx_completion_hist[idx]);
		} else {
			seq_printf(m, "%u: %u%c", idx, alsa_data->tx_completion_hist[idx],
				(idx && !((idx + 1) % 5) ? '\n' : '\t'));
		}
	}
	if (alsa_data->tx_completion_max > 0) {
		seq_printf(m, "TX Completion high score: %u\n", alsa_data->tx_completion_max);
	}
	if (rx_runtime) {
		seq_printf(m, "Current RX DMA location in frames: %08lu\n", rx_runtime->status->hw_ptr);
	}
	seq_printf(m, "Last completed RX frame: %08llu\n", alsa_data->last_rx_frame);
	seq_printf(m, "RX Buffer completion histogram:\n");
	for (idx = 0; idx < COMPLETION_HIST_SIZE; idx++) {
		if (idx == COMPLETION_HIST_SIZE - 1) {
			seq_printf(m, "%u+: %u\n", idx, alsa_data->rx_completion_hist[idx]);
		} else {
			seq_printf(m, "%u: %u%c", idx, alsa_data->rx_completion_hist[idx],
				(idx && !((idx + 1) % 5) ? '\n' : '\t'));
		}
	}
	if (alsa_data->rx_completion_max > 0) {
		seq_printf(m, "RX Completion high score: %u\n", alsa_data->rx_completion_max);
	}

	return 0;
}

static int alsa_procfs_open(struct inode *inode, struct file *file)
{
	return single_open(file, alsa_procfs_show, PDE_DATA(inode));
}

static const struct file_operations procfs_alsa_ops = {
	.owner = THIS_MODULE,
	.open = alsa_procfs_open,
	.read = seq_read,
	.llseek = seq_lseek,
	.release = single_release,
};

static int alsa_procfs_init(struct alsa_lla *alsa_data)
{
	struct proc_dir_entry *file;
	file = proc_create_data("driver/alsa_lla", S_IRUGO, NULL, &procfs_alsa_ops, alsa_data);
	if (!file) {
		bb_log_dev(alsa_data->dev, BB_MOD_LLA, BB_LVL_ERR, "Failed to create procfs file.");
		return -1;
	}
#ifdef SNOOP_RX
	file = proc_create_data("driver/alsa_snoop", S_IRUGO, NULL, &procfs_snoop_alsa_ops, alsa_data);
	if (!file) {
		bb_log_dev(alsa_data->dev, BB_MOD_LLA, BB_LVL_ERR, "Failed to create procfs file.");
		return -1;
	}
#endif
	return 0;
}

static void alsa_procfs_exit(void)
{
#ifdef SNOOP_RX
	remove_proc_entry("driver/alsa_snoop", NULL);
#endif
	remove_proc_entry("driver/alsa_lla", NULL);
}

static int alsa_lla_probe(struct platform_device *pdev)
{
	struct alsa_lla *alsa_data;
	int ret = 0;

	bb_log_dev(&pdev->dev, BB_MOD_LLA, BB_LVL_INFO, "alsa probe");
	alsa_data = devm_kzalloc(&pdev->dev, sizeof(struct alsa_lla), GFP_KERNEL);
	if (!alsa_data) {
		bb_log_dev(&pdev->dev, BB_MOD_LLA, BB_LVL_ERR, "Memory allocation failed.");
		return -ENOMEM;
	}
	platform_set_drvdata(pdev, alsa_data);
	alsa_data->dev = &(pdev->dev);
	global_data = alsa_data;

	ret = of_property_read_u32(pdev->dev.of_node, "tx_mask", &alsa_data->tx_mask);
	if (ret < 0) {
		bb_log_dev(alsa_data->dev, BB_MOD_LLA, BB_LVL_ERR, "Failed to get TX mask %d.", ret);
		return ret;
	}
	ret = of_property_read_u32(pdev->dev.of_node, "rx_mask", &alsa_data->rx_mask);
	if (ret < 0) {
		bb_log_dev(alsa_data->dev, BB_MOD_LLA, BB_LVL_ERR, "Failed to get RX mask %d.", ret);
		return ret;
	}

	init_waitqueue_head(&(alsa_data->tx_poll_event));
	init_waitqueue_head(&(alsa_data->rx_poll_event));

	lla_init_dev(&(alsa_data->tx_dev), alsa_data->dev, "alsa-tx", LLA_DEV_TYPE_DAC, &alsa_lla_tx_ops, alsa_data);
	lla_init_dev(&(alsa_data->rx_dev), alsa_data->dev, "alsa-rx", LLA_DEV_TYPE_MIC, &alsa_lla_rx_ops, alsa_data);
	ret = snd_pcm_notify(&alsa_lla_notify, 0);
	if (ret < 0) {
		bb_log_dev(alsa_data->dev, BB_MOD_LLA, BB_LVL_ERR, "Failed to register ALSA callbacks %d.", ret);
		return ret;
	}

	alsa_procfs_init(alsa_data);

	return ret;
}

static int alsa_lla_remove(struct platform_device *pdev)
{
	int ret = 0;

	alsa_procfs_exit();
	ret = snd_pcm_notify(&alsa_lla_notify, 1);
	if (ret < 0) {
		bb_log_dev(&pdev->dev, BB_MOD_LLA, BB_LVL_ERR, "Failed to unregister ALSA callbacks %d.", ret);
	}

	return ret;
}

static int alsa_lla_suspend(struct platform_device *pdev, pm_message_t state)
{
	if (global_data && global_data->dev == &(pdev->dev)) {
		if (global_data->tx_shared) {
			bb_log(BB_MOD_LLA, BB_LVL_INFO, "Clear tx LLA_SHARED_FLAG_RUNNING prior to suspend");
			global_data->tx_shared->flags &= ~LLA_SHARED_FLAG_RUNNING;
			global_data->running_flag_cleared_on_suspend |= TX_RUNNING_FLAG_CLEARED;
		}
		if (global_data->rx_shared) {
			bb_log(BB_MOD_LLA, BB_LVL_INFO, "Clear rx LLA_SHARED_FLAG_RUNNING prior to suspend");
			global_data->rx_shared->flags &= ~LLA_SHARED_FLAG_RUNNING;
			global_data->running_flag_cleared_on_suspend |= RX_RUNNING_FLAG_CLEARED;
		}
	}
	return 0;
}

static int alsa_lla_resume(struct platform_device *pdev)
{
	if (global_data && global_data->dev == &(pdev->dev)) {
		if (global_data->tx_shared) {
			if (global_data->running_flag_cleared_on_suspend & TX_RUNNING_FLAG_CLEARED) {
				bb_log(BB_MOD_LLA, BB_LVL_INFO, "Setting tx LLA_SHARED_FLAG_RUNNING and LLA_SHARED_FLAG_RESTARTED");
				global_data->tx_shared->flags |= (LLA_SHARED_FLAG_RUNNING | LLA_SHARED_FLAG_RESTARTED);
			}
		}
		if (global_data->rx_shared) {
			if (global_data->running_flag_cleared_on_suspend & RX_RUNNING_FLAG_CLEARED) {
				bb_log(BB_MOD_LLA, BB_LVL_INFO, "Setting rx LLA_SHARED_FLAG_RUNNING and LLA_SHARED_FLAG_RESTARTED");
				global_data->rx_shared->flags |= (LLA_SHARED_FLAG_RUNNING | LLA_SHARED_FLAG_RESTARTED);
			}
		}
		global_data->running_flag_cleared_on_suspend = 0;
	}
	return 0;
}

static const struct of_device_id alsa_lla_ids[] = {
	{ .compatible = "sonos,alsa-lla", },
	{}
};

static struct platform_driver alsa_lla_driver = {
	.probe = alsa_lla_probe,
	.remove = alsa_lla_remove,
	.suspend = alsa_lla_suspend,
	.resume = alsa_lla_resume,
	.driver = {
		.name = DRV_NAME,
		.owner = THIS_MODULE,
		.of_match_table = alsa_lla_ids,
	},
};

int alsa_lla_init(void) {
	bb_log(BB_MOD_LLA, BB_LVL_INFO, "alsa init");
	return platform_driver_register(&alsa_lla_driver);
}

void alsa_lla_exit(void) {
	platform_driver_unregister(&alsa_lla_driver);
}
