/* Copyright (c) 2018, Sonos, Inc.  All rights reserved. */

#include <linux/slab.h>
#include <linux/sched.h>
#include <linux/sonos_kernel.h>
#include <asm/uaccess.h>
#include "power_button_event.h"

#include "blackbox.h"
#include "power_event_queue_api.h"

struct power_button_event_queue * power_button_event_queue_create(u32 len, wait_queue_head_t *wq, spinlock_t *lock)
{
    size_t size;
    struct power_button_event_queue *beq;

    size = sizeof(*beq) + (len * sizeof(beq->events));
    beq = kmalloc(size, GFP_KERNEL);
    if (beq == NULL) {
#if defined (SONOS_ARCH_ATTR_SOC_IS_A113)
	if (PRODUCT_ID_IS_DHUEZ)  {
		bb_log(BB_MOD_EVENT, BB_LVL_ERR, "Power Event queue allocation failed!");
	}
#else
        printk("%s: failed\n", __func__);
#endif
    } else {
        memset(beq, 0, sizeof(*beq));
        beq->len = len;
        beq->kwq = wq;
        beq->lock = lock;
    }
    return beq;
}

void power_button_event_queue_free(struct power_button_event_queue *beq)
{
    if (beq != NULL) {
        kfree(beq);
    }
}

int power_button_event_send(struct power_button_event_queue *beq, enum PowerEventSource in_source, enum PowerEventInfo in_info)
{
    struct power_button_evt *event;
    enum PowerEventSource source = in_source;
    enum PowerEventInfo info = in_info;
    struct timeval timestamp;
    unsigned long flags;
    int ret = 0;

    if (beq == NULL) {
        return -EIO;
    }

#if defined (SONOS_ARCH_ATTR_SOC_IS_A113)
    if (PRODUCT_ID_IS_DHUEZ)  {
        bb_log(BB_MOD_EVENT, BB_LVL_DEBUG, "Sent event %s from %s.", power_event_queue_get_event_string(in_info),
                                                                     power_event_queue_get_source_string(in_source));
    }
#endif
    do_gettimeofday(&timestamp);
    spin_lock_irqsave(beq->lock, flags);
    event = &beq->events[beq->head];
    if (power_button_event_queue_full(beq)) {
        beq->overruns++;
        if (!beq->was_full) {
            beq->was_full = 1;
            printk("event dropped, button queue full\n");
        }
        power_button_event_queue_flush(beq);
        event = &beq->events[beq->head];
        source = POWER_EVTSOURCE_NO_SOURCE;
        info = POWER_EVTINFO_QUEUE_OVERRUN;
        ret = -EIO;
    }
    event->m_lSource   = source;
    event->m_lInfo     = info;
    event->m_timestamp = timestamp;
    beq->head = power_button_event_queue_next_head(beq);
    beq->num++;
    if (beq->num > beq->high_water) {
        beq->high_water = beq->num;
    }
    spin_unlock_irqrestore(beq->lock, flags);
    wake_up_interruptible(beq->kwq);
    return 0;
}

struct power_button_evt * power_button_event_get(struct power_button_event_queue *beq)
{
    struct power_button_evt *event = &beq->events[beq->tail];
    beq->tail = power_button_event_queue_next_tail(beq);
    beq->num--;
    return event;
}

int power_button_event_receive(struct power_button_event_queue *beq, char *user_data)
{
    struct power_button_evt *event;
    int ret = 0;
    wait_queue_t wq;

    if (beq == NULL) {
        return -EIO;
    }
    init_waitqueue_entry(&wq, current);
    spin_lock_irq(beq->lock);
    while (power_button_event_queue_empty(beq)) {
        add_wait_queue(beq->kwq, &wq);
        spin_unlock_irq(beq->lock);
        current->state = TASK_INTERRUPTIBLE;
        schedule();
        current->state = TASK_RUNNING;
        spin_lock_irq(beq->lock);
        remove_wait_queue(beq->kwq, &wq);
        if (signal_pending(current)) {
            ret = -ERESTARTSYS;
            goto out_unlock;
        }
    }

    event = power_button_event_get(beq);
    if (copy_to_user(user_data, (char *)event, sizeof(*event))) {
#if defined (SONOS_ARCH_ATTR_SOC_IS_A113)
	if (PRODUCT_ID_IS_DHUEZ)  {
		bb_log(BB_MOD_EVENT, BB_LVL_ERR, "%s: copy_to_user failed", __func__);
	}
#else
        printk("%s: copy_to_user failed\n", __func__);
#endif
        ret = -EFAULT;
    } else {
        ret = sizeof(*event);
        beq->was_full = 0;
    }

out_unlock:
    spin_unlock_irq(beq->lock);
    return ret;
}

void power_button_event_queue_flush(struct power_button_event_queue *beq)
{
    if (beq != NULL) {
        beq->head = beq->tail;
        beq->num = 0;
    }
}
