/*
 * Copyright (c) 2018, Sonos, Inc.
 * All rights reserved.
 */

#include <linux/autoconf.h>
#include <linux/kernel.h>
#include <linux/export.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/barrier.h>
#include <linux/spinlock.h>
#include <linux/circ_buf.h>
#include <linux/watchdog.h>
#include <linux/semaphore.h>
#include <linux/proc_fs.h>
#include <linux/init.h>
#include <linux/poll.h>
#include <linux/hardirq.h>
#include "sonos_device.h"
#include "blackbox.h"
#include "power_event_queue.h"
#include "power_event_queue_hal.h"

struct power_event_queue_deferred_event {
	struct work_struct worker;
	enum PowerEventSource source;
	enum PowerEventInfo info;
};

#define POWER_EVENT_QUEUE_NUM_DEFERRED_EVENTS  32
struct power_event_queue_deferred_event power_event_queue_deferred_events[POWER_EVENT_QUEUE_NUM_DEFERRED_EVENTS];
struct circ_buf power_event_queue_deferred_events_cb;
struct workqueue_struct	*power_event_queue_deferred_queue;
static spinlock_t power_event_queue_producer_lock;
static spinlock_t power_event_queue_consumer_lock;

static int devno = -1;

static struct power_button_event_queue *power_event_queue_default = NULL;

static wait_queue_head_t power_event_queue_wait_queue;
static spinlock_t power_event_queue_lock;

struct file_operations power_event_queue_fops;

static int power_device_init_success = 0;
static int power_device_readers = 0;
static int power_device_writers = 0;

static void power_event_queue_worker(struct work_struct *d)
{
	struct power_event_queue_deferred_event *decb = container_of(d, struct power_event_queue_deferred_event, worker);
	struct power_event_queue_deferred_event de;
	unsigned long head, tail;
	int power_event_present = 1;

	spin_lock(&power_event_queue_consumer_lock);
	head = power_event_queue_deferred_events_cb.head;
	tail = ACCESS_ONCE(power_event_queue_deferred_events_cb.tail);
	if (CIRC_CNT(head, tail, POWER_EVENT_QUEUE_NUM_DEFERRED_EVENTS) >= 1) {
		smp_read_barrier_depends();
		de = *decb;
		smp_mb();
		power_event_queue_deferred_events_cb.tail = (tail + 1) & (POWER_EVENT_QUEUE_NUM_DEFERRED_EVENTS - 1);
	} else {
		bb_log(BB_MOD_EVENT, BB_LVL_ERR, "%s: event missing", __func__);
		power_event_present = 0;
	}
	spin_unlock(&power_event_queue_consumer_lock);

	if (power_event_present) {
		power_event_queue_process_regs(de.source, de.info);
	}
}

enum PowerEventInfo power_event_queue_last_info[POWER_EVTSOURCE_NUM_SOURCES];

static void power_event_queue_init_events(void)
{
	enum PowerEventSource source;
	for (source = POWER_EVTSOURCE_NO_SOURCE; source < POWER_EVTSOURCE_NUM_SOURCES; source++) {
		power_event_queue_last_info[source] = POWER_EVTINFO_NO_EVENT;
	}
}

enum PowerEventInfo power_event_queue_get_event(enum PowerEventSource source)
{
	if (source >= POWER_EVTSOURCE_NUM_SOURCES) {
		return POWER_EVTINFO_NO_EVENT;
	}
	return power_event_queue_last_info[source];
}
EXPORT_SYMBOL(power_event_queue_get_event);

int power_event_queue_send_event(enum PowerEventSource source, enum PowerEventInfo info)
{
	power_event_queue_last_info[source] = info;
	power_event_queue_process_regs(source, info);
	return power_button_event_send(power_event_queue_default, source, info);
}
EXPORT_SYMBOL(power_event_queue_send_event);

int power_event_queue_send_event_defer(enum PowerEventSource source, enum PowerEventInfo info)
{
	int error, success;
	unsigned long flags, head, tail;

	error = power_button_event_send(power_event_queue_default, source, info);
	if (error) {
		bb_log(BB_MOD_EVENT, BB_LVL_ERR, "%s: event couldn't be sent, error %d", __func__, error);
		return error;
	}

	spin_lock_irqsave(&power_event_queue_producer_lock, flags);
	head = power_event_queue_deferred_events_cb.head;
	tail = ACCESS_ONCE(power_event_queue_deferred_events_cb.tail);
	if (CIRC_SPACE(head, tail, POWER_EVENT_QUEUE_NUM_DEFERRED_EVENTS) >= 1) {
		struct power_event_queue_deferred_event *de = &power_event_queue_deferred_events[head];
		INIT_WORK(&de->worker, power_event_queue_worker);
		de->source = source;
		de->info = info;
		smp_wmb();
		power_event_queue_deferred_events_cb.head = (head + 1) & (POWER_EVENT_QUEUE_NUM_DEFERRED_EVENTS - 1);
		success = queue_work(power_event_queue_deferred_queue, &de->worker);
		if (!success) {
			bb_log(BB_MOD_EVENT, BB_LVL_ERR, "%s: error queuing work", __func__);
			error = -EIO;
		}
	} else {
		bb_log(BB_MOD_EVENT, BB_LVL_ERR, "%s: no space for event", __func__);
		error = -EIO;
	}
	spin_unlock_irqrestore(&power_event_queue_producer_lock, flags);
	return error;
}
EXPORT_SYMBOL(power_event_queue_send_event_defer);

struct cdev power_event_queue_chr_dev;

static int __init power_event_queue_init(void)
{
	int error;
	struct device *class_dev = NULL;

	devno = MKDEV(POWER_EVENT_QUEUE_MAJOR_NUMBER, 0);

	power_event_queue_deferred_queue = create_singlethread_workqueue("deferred power events");
	if (power_event_queue_deferred_queue == NULL) {
		bb_log(BB_MOD_EVENT, BB_LVL_ERR, "deferred power event queue could not be created");
		return -EIO;
	}
	power_event_queue_init_events();

#ifdef CONFIG_DEVTMPFS
	error = alloc_chrdev_region(&devno, 0, 1, POWER_EVENT_QUEUE_DEVICE_NAME);
#else
	error = register_chrdev_region(devno, 1, POWER_EVENT_QUEUE_DEVICE_NAME);
#endif
	if (error) {
		bb_log(BB_MOD_EVENT, BB_LVL_ERR, "Error %d registering device %s", error, POWER_EVENT_QUEUE_DEVICE_NAME);
		goto event_queue_fail1;
	}
	cdev_init(&power_event_queue_chr_dev, &power_event_queue_fops);
	power_event_queue_chr_dev.owner = THIS_MODULE;

	error = cdev_add(&power_event_queue_chr_dev, devno, 1);
	if (error) {
		printk("Couldn't add power character device (%d).", error);
		goto event_queue_fail2;
	}

	class_dev = sonos_device_create(NULL, devno, NULL, POWER_EVENT_QUEUE_DEVICE_NAME);
	if (IS_ERR(class_dev)) {
		bb_log(BB_MOD_EVENT, BB_LVL_ERR, "Error creating event queue class.\n");
		cdev_del(&power_event_queue_chr_dev);
		error = PTR_ERR(class_dev);
		goto event_queue_fail2;
	}

	spin_lock_init(&power_event_queue_lock);
	spin_lock_init(&power_event_queue_producer_lock);
	spin_lock_init(&power_event_queue_consumer_lock);
	init_waitqueue_head(&power_event_queue_wait_queue);
	power_event_queue_default = power_button_event_queue_create(100, &power_event_queue_wait_queue, &power_event_queue_lock);
	if (power_event_queue_default == NULL) {
		bb_log(BB_MOD_EVENT, BB_LVL_ERR, "default event power queue couldn't be created");
		goto event_queue_fail3;
	}
	error = power_event_queue_init_proc();
	if (error) {
		bb_log(BB_MOD_EVENT, BB_LVL_ERR, "event power queue proc init failed, error %d", error);
		goto event_queue_fail4;
	}
	power_event_queue_deferred_events_cb.head = 0;
	power_event_queue_deferred_events_cb.tail = 0;
	bb_log(BB_MOD_EVENT, BB_LVL_INFO, "Power Event queue registered at %d, version %d", MAJOR(devno), POWER_EVENT_QUEUE_VERSION);
	power_device_init_success = 1;
	return 0;

event_queue_fail4:
	power_button_event_queue_free(power_event_queue_default);
	power_event_queue_default = NULL;

event_queue_fail3:
	sonos_device_destroy(devno);
	cdev_del(&power_event_queue_chr_dev);

event_queue_fail2:
	unregister_chrdev_region(devno, 1);

event_queue_fail1:
	destroy_workqueue(power_event_queue_deferred_queue);

	return error;
}
module_init(power_event_queue_init);

static void __exit power_event_queue_exit(void)
{
	power_event_queue_remove_proc();
	power_button_event_queue_free(power_event_queue_default);
	power_event_queue_default = NULL;
	sonos_device_destroy(devno);
	cdev_del(&power_event_queue_chr_dev);
	unregister_chrdev_region(devno, 1);
	destroy_workqueue(power_event_queue_deferred_queue);
}
module_exit(power_event_queue_exit);

static int power_event_queue_device_open(struct inode *inode, struct file *file)
{

	if (file->f_mode & FMODE_READ) {
		if (!power_device_readers++) {
			power_button_event_queue_flush(power_event_queue_default);
		}
	}

	if (file->f_mode & FMODE_WRITE) {
		++power_device_writers;
	}

	return (power_device_init_success ? 0 : -ENODEV);
};

static int power_event_queue_device_release(struct inode *inode, struct file *file)
{
	if (file->f_mode & FMODE_READ) {
		--power_device_readers;
	}

	if (file->f_mode & FMODE_WRITE) {
		--power_device_writers;
	}

	return 0;
}

static ssize_t power_event_queue_device_read(struct file *filp, char *data, size_t len, loff_t *ppos)
{
	int ret;

	if (len < sizeof(struct power_button_evt)) {
		return -EINVAL;
	}

	if (filp->f_flags & O_NONBLOCK) {
		if (power_button_event_queue_empty(power_event_queue_default)) {
			return -EWOULDBLOCK;
		}
	}

	ret = power_button_event_receive(power_event_queue_default, data);
	if (ret < 0) {
		return ret;
	}
	return sizeof(struct power_button_evt);
}

static ssize_t power_event_queue_device_write(struct file *file, const char *data, size_t len, loff_t *ppos)
{
	return 0;
}

static unsigned int power_event_queue_device_poll(struct file *fp, struct poll_table_struct *pt)
{
	unsigned int mask=0;

	poll_wait(fp,&power_event_queue_wait_queue, pt);
	if (!power_button_event_queue_empty(power_event_queue_default)) {
		mask = POLLIN | POLLRDNORM;
	}

	return mask;
}

static long power_event_queue_device_ioctl(struct file *filp, unsigned int cmd, unsigned long arg)
{
	switch (_IOC_NR(cmd)) {
	case 0:
	{
		uint32_t ver = POWER_EVENT_QUEUE_VERSION;
		if (copy_to_user((uint32_t *)arg, &ver, sizeof(uint32_t))) {
			return -EACCES;
		}
		break;
	}
	case 12:
	case 28:
	case 58:
	case 61:
	case 65:
	case 66:
	case 68:
		break;
	default:
		printk("power_event_queue: Unrecognized IOCTL command %u\n", _IOC_NR(cmd));
		return -EINVAL;
	}
	return 0;
}

struct file_operations power_event_queue_fops =
{
	.unlocked_ioctl = power_event_queue_device_ioctl,
	.open		= power_event_queue_device_open,
	.release	= power_event_queue_device_release,
	.read		= power_event_queue_device_read,
	.write		= power_event_queue_device_write,
	.poll		= power_event_queue_device_poll,
};

MODULE_LICENSE ("GPL");
MODULE_AUTHOR ("Sonos, Inc.");
MODULE_DESCRIPTION ("Power Event Queue Module");
