/*
 * Copyright (c) 2018 Sonos Inc.
 * All rights reserved.
 */

#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/hardirq.h>
#include <linux/rwsem.h>
#include "blackbox.h"
#include "power_event_queue.h"


#define POWER_EVENT_REG_MAGIC		0xdef1ade4


struct list_head power_event_queue_reg_list = LIST_HEAD_INIT(power_event_queue_reg_list);
DECLARE_RWSEM(power_event_queue_reg_lock);

static inline int power_eqr_match_source(struct power_event_reg *eqr, enum PowerEventSource source)
{
	if ((eqr->source == POWER_EVTSOURCE_NUM_SOURCES) || (eqr->source == source)) {
		return 1;
	}
	return 0;
}

static inline int power_eqr_match_info(struct power_event_reg *eqr, enum PowerEventInfo info)
{
	if ((eqr->info == POWER_EVTINFO_NUM_EVENTS) || (eqr->info == info)) {
		return 1;
	}
	return 0;
}

static inline int power_eqr_match_callback(struct power_event_reg *eqr, power_event_cbfunc cbfunc)
{
	return (eqr->cbfunc == cbfunc);
}

static inline int power_eqr_match_param(struct power_event_reg *eqr, void *param)
{
	return (eqr->param == param);
}

struct power_event_reg *
power_event_register_event(enum PowerEventSource source, enum PowerEventInfo info, power_event_cbfunc cb, void *param, char *name)
{
	struct power_event_reg *eqr;
	int found = 0;

	down_read(&power_event_queue_reg_lock);
	list_for_each_entry(eqr, &power_event_queue_reg_list, list)
	{
		if (power_eqr_match_source(eqr, source) &&
			power_eqr_match_info(eqr, info) &&
			power_eqr_match_callback(eqr, cb) &&
			power_eqr_match_param(eqr, param)) {
			found = 1;
			break;
		}
	}
	up_read(&power_event_queue_reg_lock);
	if (found) {
		goto event_register_exit;
	}

	eqr = kmalloc(sizeof(*eqr), GFP_KERNEL);
	if (eqr == NULL) {
		bb_log(BB_MOD_EVENT, BB_LVL_ERR, "%s: unable to alloc EQR", __func__);
		return NULL;
	}
	eqr->param = param;
	eqr->cbfunc = cb;
	eqr->source = source;
	eqr->info = info;
	eqr->magic = POWER_EVENT_REG_MAGIC;
	strncpy(eqr->name, name, POWER_EVENT_REG_NAME_LEN);
	eqr->name[POWER_EVENT_REG_NAME_LEN-1] = '\0';
	down_write(&power_event_queue_reg_lock);
	list_add(&eqr->list, &power_event_queue_reg_list);
	up_write(&power_event_queue_reg_lock);

event_register_exit:
	return eqr;
}
EXPORT_SYMBOL(power_event_register_event);

struct power_event_reg *
power_event_register_source(enum PowerEventSource source, power_event_cbfunc cb, void *param, char *name)
{
	return power_event_register_event(source, POWER_EVTINFO_NUM_EVENTS, cb, param, name);
}
EXPORT_SYMBOL(power_event_register_source);

struct power_event_reg *
power_event_register_info(enum PowerEventInfo info, power_event_cbfunc cb, void *param, char *name)
{
	return power_event_register_event(POWER_EVTSOURCE_NUM_SOURCES, info, cb, param, name);
}
EXPORT_SYMBOL(power_event_register_info);

struct power_event_reg *
power_event_register_all(power_event_cbfunc cb, void *param, char *name)
{
	return power_event_register_event(POWER_EVTSOURCE_NUM_SOURCES, POWER_EVTINFO_NUM_EVENTS, cb, param, name);
}
EXPORT_SYMBOL(power_event_register_all);

int power_event_unregister_event(struct power_event_reg **peqr)
{
	int error = 0;
	struct power_event_reg *eqr = *peqr;
	if (eqr != NULL) {
		if (eqr->magic == POWER_EVENT_REG_MAGIC) {
			list_del(&eqr->list);
			eqr->magic = 0;
			kfree(eqr);
			*peqr = NULL;
		} else {
			error = -1;
		}
	}
	return error;
}
EXPORT_SYMBOL(power_event_unregister_event);

void power_event_queue_process_regs(enum PowerEventSource source, enum PowerEventInfo info)
{
	struct power_event_reg *eqr;

	if (in_atomic()) {
		panic("power event generated from illegal context, source %s event %s\n",
			  power_event_queue_get_source_string(source), power_event_queue_get_event_string(info));
	}

	down_read(&power_event_queue_reg_lock);
	list_for_each_entry(eqr, &power_event_queue_reg_list, list)
	{
		if (power_eqr_match_source(eqr, source) && power_eqr_match_info(eqr, info)) {
			eqr->cbfunc(eqr->param, source, info);
			eqr->calls++;
		}
	}
	up_read(&power_event_queue_reg_lock);
}
